@@ -18,6 +18,7 @@ from typing import (
     Iterator,
     List,
     Optional,
+    Set,
     Union,
     cast,
 )
@@ -30,26 +31,32 @@ import discord.errors
 import hupper  # type: ignore
 import requests
 from discord.channel import TextChannel
+from dotenv import load_dotenv
 from sqlalchemy import exc
 from sqlalchemy.orm.session import Session
 from sqlalchemy.sql import text
 
 from spellbot._version import __version__
 from spellbot.assets import ASSET_FILES, s
-from spellbot.constants import ADMIN_ROLE, CREATE_ENDPOINT, DEFAULT_GAME_SIZE, THUMB_URL
+from spellbot.constants import (
+    ADMIN_ROLE,
+    CREATE_ENDPOINT,
+    DEFAULT_GAME_SIZE,
+    INVITE_LINK,
+    THUMB_URL,
+)
 from spellbot.data import Channel, Data, Event, Game, Server, Tag, User
 
+load_dotenv()
+
 # Application Paths
 RUNTIME_ROOT = Path(".")
 SCRIPTS_DIR = RUNTIME_ROOT / "scripts"
 DB_DIR = RUNTIME_ROOT / "db"
-DEFAULT_DB_URL = f"sqlite:///{DB_DIR}/spellbot.db"
 TMP_DIR = RUNTIME_ROOT / "tmp"
 MIGRATIONS_DIR = SCRIPTS_DIR / "migrations"
-INVITE_LINK = (
-    "https://discordapp.com/api/oauth2/authorize"
-    "?client_id=725510263251402832&permissions=92224&scope=bot"
-)
+
+DEFAULT_DB_URL = f"sqlite:///{DB_DIR}/spellbot.db"
 
 logger = logging.getLogger(__name__)
 
@@ -160,6 +167,10 @@ def parse_opts(params: List[str]) -> dict:
     }
 
 
+def post_link(server_xid: int, channel_xid: int, message_xid: int) -> str:
+    return f"https://discordapp.com/channels/{server_xid}/{channel_xid}/{message_xid}"
+
+
 def is_admin(
     channel: TextChannel, user_or_member: Union[discord.User, discord.Member]
 ) -> bool:
@@ -388,7 +399,9 @@ class SpellBot(discord.Client):
         self, loop: asyncio.AbstractEventLoop
     ) -> None:  # pragma: no cover
         """Start up any periodic background tasks."""
-        self.cleanup_expired_games_task(loop)
+        # TODO: Think about how to make this better...
+        # self.cleanup_expired_games_task(loop)
+
         # TODO: Make this manually triggered.
         # self.cleanup_started_games_task(loop)
 
@@ -670,7 +683,7 @@ class SpellBot(discord.Client):
                 session.query(Game).filter(Game.message_xid == message.id).one_or_none()
             )
             if not game or game.status != "pending":
-                return  # this isn't a post relating to a game, just ignore it
+                return  # this isn't a post relating to a pending game, just ignore it
 
             user = self.ensure_user_exists(session, author)
 
@@ -683,16 +696,16 @@ class SpellBot(discord.Client):
                 if any(user.xid == game_user.xid for game_user in game.users):
                     # this author is already in this game, they don't need to be added
                     return
-                if user.game and user.game.id != game.id:
-                    # this author is already another game, they can't be added
-                    await author.send(
-                        s(
-                            "react_already_in",
-                            reply=f"<@{author.id}>",
-                            prefix=server.prefix,
-                        )
-                    )
-                    return
+                if (
+                    user.game
+                    and user.game.id != game.id
+                    and user.game.status != "started"
+                ):
+                    # this author is already another game, leave that one now
+                    game_to_update = user.game
+                    user.game_id = None
+                    session.commit()
+                    await self.try_to_update_game(game_to_update)
                 user.game = game
             else:  # emoji == "➖":
                 if not any(user.xid == game_user.xid for game_user in game.users):
@@ -736,42 +749,49 @@ class SpellBot(discord.Client):
 
     async def on_message(self, message: discord.Message) -> None:
         """Behavior when the client gets a message from Discord."""
-        # don't respond to any bots
-        if cast(discord.User, message.author).bot:
-            return
+        try:
+            author = cast(discord.User, message.author)
 
-        private = str(message.channel.type) == "private"
+            # don't respond to any bots
+            if author.bot:
+                return
 
-        # only respond in text channels and to direct messages
-        if not private and str(message.channel.type) != "text":
-            return
+            private = str(message.channel.type) == "private"
 
-        # handle confirm/deny invitations over direct message
-        if private:
-            choice = message.content.lstrip()[0:3].rstrip().lower()
-            if choice in ["yes", "no"]:
-                return await self.process_invite_response(message, choice)
+            # only respond in text channels and to direct messages
+            if not private and str(message.channel.type) != "text":
+                return
 
-        # only respond to command-like messages
-        if not private:
-            rows = self.data.conn.execute(
-                text("SELECT prefix FROM servers WHERE guild_xid = :g"),
-                g=message.channel.guild.id,
-            )
-            prefixes = [row.prefix for row in rows]
-            prefix = prefixes[0] if prefixes else "!"
-        else:
-            prefix = "!"
-        if not message.content.startswith(prefix):
-            return
+            # handle confirm/deny invitations over direct message
+            if private:
+                choice = message.content.lstrip()[0:3].rstrip().lower()
+                if choice in ["yes", "no"]:
+                    return await self.process_invite_response(message, choice)
 
-        if not private:
-            async with self.session() as session:
-                server = self.ensure_server_exists(session, message.channel.guild.id)
-                if not server.bot_allowed_in(message.channel.id):
-                    return
+            # only respond to command-like messages
+            if not private:
+                rows = self.data.conn.execute(
+                    text("SELECT prefix FROM servers WHERE guild_xid = :g"),
+                    g=message.channel.guild.id,
+                )
+                prefixes = [row.prefix for row in rows]
+                prefix = prefixes[0] if prefixes else "!"
+            else:
+                prefix = "!"
+            if not message.content.startswith(prefix):
+                return
 
-        await self.process(message, prefix)
+            is_admin = author.permissions_in(message.channel).administrator
+            if not private and not is_admin:
+                async with self.session() as session:
+                    server = self.ensure_server_exists(session, message.channel.guild.id)
+                    if not server.bot_allowed_in(message.channel.id):
+                        return
+
+            await self.process(message, prefix)
+        except Exception as e:
+            logging.exception("unhandled exception:", e)
+            raise
 
     async def on_ready(self) -> None:
         """Behavior when the client has successfully connected to Discord."""
@@ -837,21 +857,22 @@ class SpellBot(discord.Client):
             usage += f"\n`{title}`"
             usage += f"\n>  {use}"
             usage += "\n"
-        usage += "---"
+        usage += "---\n"
         usage += (
-            " \nPlease report any bugs and suggestions at"
-            " <https://github.com/lexicalunit/spellbot/issues>!"
+            "Please report any bugs and suggestions at"
+            " <https://github.com/lexicalunit/spellbot/issues>!\n"
         )
         usage += "\n"
-        usage += f"[🔗 Add SpellBot to your Discord!]({INVITE_LINK})\n"
+        usage += f"Add SpellBot to your Discord: <{INVITE_LINK}>\n"
         usage += "\n"
         usage += (
-            "\n💜 You can help keep SpellBot running by supporting me on Ko-fi! "
+            "💜 You can help keep SpellBot running by supporting me on Ko-fi! "
             "<https://ko-fi.com/Y8Y51VTHZ>"
         )
-        await message.channel.send(
-            s("dm_sent", reply=f"<@{cast(discord.User, message.author).id}>")
-        )
+        if str(message.channel.type) != "private":
+            await message.channel.send(
+                s("dm_sent", reply=f"<@{cast(discord.User, message.author).id}>")
+            )
         for page in paginate(usage):
             await message.author.send(page)
 
@@ -866,13 +887,10 @@ class SpellBot(discord.Client):
         embed.set_thumbnail(url=THUMB_URL)
         version = f"[{__version__}](https://pypi.org/project/spellbot/{__version__}/)"
         embed.add_field(name="Version", value=version)
-        embed.add_field(
-            name="Package", value="[PyPI](https://pypi.org/project/spellbot/)"
-        )
         author = "[@lexicalunit](https://github.com/lexicalunit)"
         embed.add_field(name="Author", value=author)
         embed.description = (
-            "_A Discord bot for [SpellTable](https://www.spelltable.com/)._\n"
+            "_The Discord bot for [SpellTable](https://www.spelltable.com/)._\n"
             "\n"
             f"Use the command `{prefix}help` for usage details. "
             "Having issues with SpellBot? "
@@ -884,7 +902,6 @@ class SpellBot(discord.Client):
             "[supporting me on Ko-fi!](https://ko-fi.com/Y8Y51VTHZ)"
         )
         embed.url = "http://spellbot.io/"
-        embed.set_footer(text="MIT © amy@lexicalunit et al")
         embed.color = discord.Color(0x5A3EFD)
         await message.channel.send(embed=embed)
 
@@ -895,19 +912,16 @@ class SpellBot(discord.Client):
         params: List[str],
         message: discord.Message,
         create_new_game: bool,
+        use_queue: bool,
     ) -> None:
         server = self.ensure_server_exists(session, message.channel.guild.id)
 
         user = self.ensure_user_exists(session, message.author)
-        if user.waiting:
-            await message.channel.send(
-                s(
-                    "user_waiting",
-                    reply=f"<@{cast(discord.User, message.author).id}>",
-                    prefix=server.prefix,
-                )
-            )
-            return
+        if not use_queue and user.waiting:
+            game_to_update = user.game
+            user.game_id = None
+            session.commit()
+            await self.try_to_update_game(game_to_update)
 
         mentions = message.mentions if message.channel.type != "private" else []
 
@@ -920,7 +934,7 @@ class SpellBot(discord.Client):
             )
             return
 
-        if len(mentions) + 1 >= size:
+        if not use_queue and len(mentions) + 1 >= size:
             await message.channel.send(
                 s(
                     "lfg_too_many_mentions",
@@ -929,22 +943,51 @@ class SpellBot(discord.Client):
             )
             return
 
+        if use_queue and len(mentions) < 1:
+            await message.channel.send(
+                s(
+                    "lfg_queue_no_mention",
+                    reply=f"<@{cast(discord.User, message.author).id}>",
+                )
+            )
+            return
+
+        if len(tag_names) > 5:
+            await message.channel.send(
+                s("tags_too_many", reply=f"<@{cast(discord.User, message.author).id}>")
+            )
+            return
+        tags = Tag.create_many(session, tag_names)
+
         game: Optional[Game] = None
         found_existing = False
         join_existing = False
 
-        mentioned_users = []
+        mentioned_users: List[User] = []
         for mentioned in mentions:
             mentioned_user = self.ensure_user_exists(session, mentioned)
             if (
-                len(mentions) == 1
+                not use_queue
+                and len(mentions) == 1
                 and mentioned_user.game
                 and mentioned_user.game.status == "pending"
             ):
+                # this user wants to join up with another player already in a game
                 game = mentioned_user.game
                 found_existing = True
                 join_existing = True
+            elif use_queue and mentioned_user.waiting:
+                # admins can pull users out of games
+                game_to_update = mentioned_user.game
+                mentioned_user.game_id = None
+                session.commit()
+                await self.try_to_update_game(game_to_update)
+                await message.channel.send(
+                    s("lfg_player_removed_from_queue", player=f"<@{mentioned_user.xid}>",)
+                )
+                return
             elif mentioned_user.waiting:
+                # at least one of the multiple players mentioend is already in a game
                 await message.channel.send(
                     s(
                         "lfg_mention_already",
@@ -955,28 +998,22 @@ class SpellBot(discord.Client):
                 return
             mentioned_users.append(mentioned_user)
 
-        if len(tag_names) > 5:
-            await message.channel.send(
-                s("tags_too_many", reply=f"<@{cast(discord.User, message.author).id}>")
-            )
-            return
-        tags = Tag.create_many(session, tag_names)
-
         now = datetime.utcnow()
         expires_at = now + timedelta(minutes=server.expire)
         user.invited = False
 
         if not game:
-            seats = 1 + len(mentions)
+            seats = 1 + len(mentions) if not use_queue else len(mentions)
             game = Game.find_existing(
                 session, server, message.channel.id, size, seats, tags, system
             )
 
         if game:
             found_existing = True
-            user.game = game
-            user.game.expires_at = expires_at
-            user.game.updated_at = now
+            if not use_queue:
+                user.game = game
+                user.game.expires_at = expires_at
+                user.game.updated_at = now
         else:
             if not create_new_game:
                 await message.channel.send(
@@ -997,20 +1034,22 @@ class SpellBot(discord.Client):
                 tags=tags,
                 server=server,
             )
-            user.game = game
+            if not use_queue:
+                user.game = game
         if not join_existing:
             for mentioned_user in mentioned_users:
                 mentioned_user.game = game
-                mentioned_user.invited = True
-                mentioned_user.invite_confirmed = False
+                if not use_queue:
+                    mentioned_user.invited = True
+                    mentioned_user.invite_confirmed = False
         session.commit()
 
         mentionor = message.author
         mentionor_xid = cast(Any, mentionor).id  # typing doesn't recognize author.id
 
-        if not join_existing:
-            if mentioned_users:
-                for mentioned in mentions:
+        if mentioned_users and not join_existing:
+            for mentioned in mentions:
+                if not use_queue:
                     await mentioned.send(
                         s(
                             "lfg_invited",
@@ -1018,12 +1057,37 @@ class SpellBot(discord.Client):
                             mention=f"<@{mentionor_xid}>",
                         )
                     )
+                else:
+                    if game.message_xid:
+                        await message.channel.send(
+                            s(
+                                "lfg_player_added_to_queue_with_link",
+                                player=f"<@{mentioned.id}>",
+                                link=post_link(
+                                    server.guild_xid, message.channel.id, game.message_xid
+                                ),
+                            )
+                        )
+                    else:
+                        await message.channel.send(
+                            s("lfg_player_added_to_queue", player=f"<@{mentioned.id}>")
+                        )
+            if not use_queue:
                 return  # we can't create the game post until we get confirmations
 
         if found_existing and game.message_xid:
             post = await self.safe_fetch_message(message.channel, game.message_xid)
             if post:
-                await message.channel.send(s("play_found", reply=f"<@{mentionor_xid}>"))
+                if not use_queue:
+                    await message.channel.send(
+                        s(
+                            "play_found",
+                            reply=f"<@{mentionor_xid}>",
+                            link=post_link(
+                                server.guild_xid, message.channel.id, game.message_xid
+                            ),
+                        )
+                    )
                 found_discord_users = []
                 if len(cast(List[User], game.users)) == game.size:
                     for game_user in game.users:
@@ -1046,13 +1110,67 @@ class SpellBot(discord.Client):
                     await self.safe_edit_message(post, embed=game.to_embed())
                 return  # Done!
 
+        if use_queue and game.size == len(mentioned_users):
+            game = Game(
+                created_at=now,
+                updated_at=now,
+                expires_at=expires_at,
+                size=size,
+                channel_xid=message.channel.id,
+                system=system,
+                url=self.create_game(),
+                status="started",
+                tags=tags,
+                server=server,
+                users=mentioned_users,
+            )
+            session.commit()
+            post = await message.channel.send(embed=game.to_embed())
+            game.message_xid = post.id
+            session.commit()
+            for mentioned_user in mentioned_users:
+                discord_user = await self.safe_fetch_user(mentioned_user.xid)
+                if discord_user:
+                    await discord_user.send(embed=game.to_embed())
+                else:
+                    pass  # TODO: Is this even possible? What should we do here...
+            return
+
         post = await message.channel.send(embed=game.to_embed())
         game.message_xid = post.id
         session.commit()
-        await post.add_reaction("➕")
-        await post.add_reaction("➖")
+        if not use_queue:
+            await post.add_reaction("➕")
+            await post.add_reaction("➖")
 
-    @command(allow_dm=False, aliases=["play", "queue"])
+    @command(allow_dm=False, aliases=["signup"])
+    async def queue(
+        self, session: Session, prefix: str, params: List[str], message: discord.Message
+    ) -> None:
+        """
+        Add a player to an admin controller play queue. _Requires the
+        "SpellBot Admin" role._
+
+        This command works in a similar way as !lfg, however it is intended to be used
+        by event runners to add players to a special game queue. For example if they have
+        paid for access to an on-demand event.
+
+        I recommend that you run this command in a channel that players don't have
+        access to view so that the game post that SpellBot makes can't be interacted
+        with by players.
+        & [size:N] [@player-1] [@player-2] ... [~tag-1] [~tag-2] ...
+        """
+        if not is_admin(message.channel, message.author):
+            await message.channel.send(
+                s("not_admin", reply=f"<@{cast(discord.User, message.author).id}>")
+            )
+            return
+
+        await self._play_helper(
+            session, prefix, params, message, create_new_game=True, use_queue=True
+        )
+
+    @command(allow_dm=False, aliases=["play"])
     async def lfg(
         self, session: Session, prefix: str, params: List[str], message: discord.Message
     ) -> None:
@@ -1075,7 +1193,9 @@ class SpellBot(discord.Client):
         are set on your game when they are looking for games to join.
         & [size:N] [@player-1] [@player-2] ... [~tag-1] [~tag-2] ...
         """
-        await self._play_helper(session, prefix, params, message, create_new_game=True)
+        await self._play_helper(
+            session, prefix, params, message, create_new_game=True, use_queue=False
+        )
 
     @command(allow_dm=False, aliases=["find", "search"])
     async def join(
@@ -1084,7 +1204,9 @@ class SpellBot(discord.Client):
         """
         Works exactly like `!lfg` except that it will NOT create a new game.
         """
-        await self._play_helper(session, prefix, params, message, create_new_game=False)
+        await self._play_helper(
+            session, prefix, params, message, create_new_game=False, use_queue=False
+        )
 
     @command(allow_dm=False)
     async def event(
@@ -1182,6 +1304,9 @@ class SpellBot(discord.Client):
         session.add(event)
         session.commit()
 
+        players_in_this_event: Set[str] = set()
+        warnings = set()
+
         members = message.channel.guild.members
         member_lookup = {member.name.lower().strip(): member for member in members}
         for i, row in enumerate(reader):
@@ -1196,12 +1321,21 @@ class SpellBot(discord.Client):
                 await message.channel.send(warning)
                 continue
 
-            warnings = set()
-
-            player_discord_users = []
+            player_discord_users: List[discord.User] = []
             for csv_data, lname in zip(csv_row_data, player_lnames):
+                if lname in players_in_this_event:
+                    await message.channel.send(
+                        s(
+                            "event_duplicate_user",
+                            row=i + 1,
+                            name=csv_data,
+                            players=players_s,
+                        )
+                    )
+                    return
                 player_discord_user = member_lookup.get(lname)
                 if player_discord_user:
+                    players_in_this_event.add(lname)
                     player_discord_users.append(player_discord_user)
                 else:
                     warnings.add(
@@ -1213,11 +1347,6 @@ class SpellBot(discord.Client):
                         )
                     )
 
-            warnings_s = "\n".join(warnings)
-            if warnings_s:
-                for page in paginate(warnings_s):
-                    await message.channel.send(page)
-
             if len(player_discord_users) != size:
                 continue
 
@@ -1254,6 +1383,18 @@ class SpellBot(discord.Client):
             session.add(game)
             session.commit()
 
+        def by_row(s: str) -> int:
+            m = re.match("^.*row ([0-9]+).*$", s)
+            # TODO: Hopefully no one adds a strings.yaml warning
+            #       that doesn't fit this exact format!
+            assert m is not None
+            return int(m[1])
+
+        warnings_s = "\n".join(sorted(warnings, key=by_row))
+        if warnings_s:
+            for page in paginate(warnings_s):
+                await message.channel.send(page)
+
         if not event.games:
             session.delete(event)
             await message.channel.send(
@@ -1336,10 +1477,11 @@ class SpellBot(discord.Client):
             game.url = self.create_game() if game.system == "spelltable" else None
             game.status = "started"
             response = game.to_embed()
+            session.commit()
+
             for discord_user in found_discord_users:
                 await discord_user.send(embed=response)
 
-            session.commit()
             await message.channel.send(
                 s(
                     "game_created",
@@ -1508,14 +1650,14 @@ class SpellBot(discord.Client):
         server = self.ensure_server_exists(session, message.channel.guild.id)
         export_file = TMP_DIR / f"{message.channel.guild.name}.csv"
         channel_name_cache = {}
+        data = server.games_data()
         with open(export_file, "w") as f, redirect_stdout(f):
             print(  # noqa: T001
-                "id,size,status,message,system,channel_xid,url,event_id,created_at,tags"
+                "id,size,status,system,channel,url,event_id,created_at,tags,message"
             )
-            data = server.games_data()
             for i in range(len(data["id"])):
                 channel_xid = data["channel_xid"][i]
-                if channel_xid not in channel_name_cache:
+                if channel_xid and channel_xid not in channel_name_cache:
                     channel = await self.safe_fetch_channel(int(channel_xid))
                     if channel:
                         name = cast(TextChannel, channel).name
@@ -1528,13 +1670,13 @@ class SpellBot(discord.Client):
                             data["id"][i],
                             data["size"][i],
                             data["status"][i],
-                            data["message"][i],
                             data["system"][i],
-                            channel_name_cache[data["channel_xid"][i]],
+                            channel_name_cache[channel_xid] if channel_xid else "",
                             data["url"][i],
                             data["event_id"][i],
                             data["created_at"][i],
                             data["tags"][i],
+                            data["message"][i],
                         ]
                     )
                 )
@@ -1612,14 +1754,33 @@ class SpellBot(discord.Client):
         # Blow away the current associations first, otherwise SQLAlchemy will explode.
         session.query(Channel).filter_by(guild_xid=server.guild_xid).delete()
 
+        all_channels = False
         channels = []
         for param in params:
+            if param.lower() == "all":
+                all_channels = True
+                break
+
             m = re.match("<#([0-9]+)>", param)
             if not m:
+                await message.channel.send(
+                    s(
+                        "spellbot_channels_warn",
+                        reply=f"<@{cast(discord.User, message.author).id}>",
+                        param=param,
+                    )
+                )
                 continue
 
             discord_channel = await self.safe_fetch_channel(int(m[1]))
             if not discord_channel:
+                await message.channel.send(
+                    s(
+                        "spellbot_channels_warn",
+                        reply=f"<@{cast(discord.User, message.author).id}>",
+                        param=param,
+                    )
+                )
                 continue
 
             channel = Channel(channel_xid=discord_channel.id, guild_xid=server.guild_xid)
@@ -1627,7 +1788,17 @@ class SpellBot(discord.Client):
             channels.append(channel)
             session.commit()
 
-        if channels:
+        if all_channels:
+            server.channels = []
+            session.commit()
+            await message.channel.send(
+                s(
+                    "spellbot_channels",
+                    reply=f"<@{cast(discord.User, message.author).id}>",
+                    channels="all channels",
+                )
+            )
+        elif channels:
             server.channels = channels
             session.commit()
             channels_str = ", ".join([f"<#{c.channel_xid}>" for c in channels])
